import asyncdispatch, cliutils, httpclient, json, os, osproc, sequtils,
  strutils, times, unittest

from langutils import sameContents

import ../testutil
import ../../../main/nim/strawbosspkg/configuration
import ../../../main/nim/strawbosspkg/server

let apiBase = "http://localhost:8180/api"
let cfgFilePath = "src/test/json/strawboss.config.json"
let cfg = loadStrawBossConfig(cfgFilePath)

let testuser = UserRef( # note: needs to correspond to an actual user
  name: "bob@builder.com",
  hashedPwd: "$2a$11$lVZ9U4optQMhzPh0E9A7Yu6XndXblUF3gCa.zmEvJy4F.4C4718b.")

suite "strawboss server":

  # suite setup code
  let serverProcess = startProcess("./strawboss", ".",
    @["serve", "-c", cfgFilePath], loadEnv(), {poUsePath})

  let http = newHttpClient()

  # give the server time to spin up
  sleep(100)

  ## UNIT TESTS

  test "validate hashed pwd":

    check validatePwd(testuser, "password")

  test "detect invalid pwds":
    check(not validatePwd(testuser, "Password"))

  test "make and extract a JWT token from a session":
    let session = newSession(testuser)
    let tok = toJWT(cfg, session)
    check fromJWT(cfg, tok) == session

  test "ping":
    let resp = http.get(apiBase & "/service/debug/ping")
    check:
      resp.status.startsWith("200")
      resp.body == "\"pong\""

  test "fail auth":
    let resp = http.post(apiBase & "/auth-token",
      $(%*{"username": "bob@builder.com", "password": "notpassword"}))
    check resp.status.startsWith("401")

  test "auth":
    let resp = http.post(apiBase & "/auth-token",
      $(%*{"username": "bob@builder.com", "password": "password"}))
    check resp.status.startsWith("200")

  test "verify valid auth token":
    let authHttp = newAuthenticatedHttpClient(apiBase, "bob@builder.com", "password")
    let resp = authHttp.get(apiBase & "/verify-auth")
    check resp.status.startsWith("200")

  test "verify fails when no auth token is given":
    let resp = http.get(apiBase & "/verify-auth")
    check resp.status.startsWith("401")

  test "verify fails when invalid auth token is given":
    let http1 = newHttpClient()
    http1.headers = newHttpHeaders({"Authorization": "Bearer nope"})
    let resp = http1.get(apiBase & "/verify-auth")
    check resp.status.startsWith("401")

  test "fail to get projects when not authenticated":
    let resp = http.get(apiBase & "/projects")
    check resp.status.startsWith("401")

  test "get projects":
    let authHttp = newAuthenticatedHttpClient(apiBase, "bob@builder.com", "password")
    let resp = authHttp.get(apiBase & "/projects")
    check resp.status.startsWith("200")

    let projects: seq[ProjectDef] = parseJson(resp.body).getElems.mapIt(parseProjectDef(it))

    check sameContents(projects, cfg.projects)

  # suite tear-down

  # give the server time to spin down but kill it after that
  discard newAsyncHttpClient().post(apiBase & "/service/debug/stop")
  sleep(100)
  if serverProcess.running: kill(serverProcess)

